// (C) Copyright Adam Wulkiewicz 2010

#ifndef SPACE_KDTREE_HPP
#define SPACE_KDTREE_HPP

// -------------------------------------------------------------------------------------- //

#include <vector>
#include <algorithm>
#include <boost/static_assert.hpp>

#include "types.hpp"

// -------------------------------------------------------------------------------------- //

namespace space
{

// -------------------------------------------------------------------------------------- //

namespace detail
{
	// -------------------------------------------------------------------------------------- //

	size_t calculate_median_index(size_t first_index, size_t last_index)
	{
		// this algorithm is designed for nodes vector where root node has index = 1
		size_t points_number = last_index - first_index;
		size_t median_index = 1;
		while ( 4 * median_index <= points_number )
			median_index += median_index;

		if ( 3 * median_index <= points_number )
		{
			median_index += median_index;
			median_index += first_index - 1;
		}
		else
			median_index = last_index - median_index;

		return median_index;
	}

	// -------------------------------------------------------------------------------------- //

	template <size_t A>
	struct pt_comp_ct
	{
		template <typename Point>
		bool operator()(Point const& p1, Point const& p2) const
		{
			return p1.template get<A>() < p2.template get<A>();
		}
	};

	// -------------------------------------------------------------------------------------- //

	template <typename PointAdaptor, typename PointCategory>
	class kdtree_algorithms
	{};

	// -------------------------------------------------------------------------------------- //

	template <typename PointAdaptor>
	class kdtree_algorithms<PointAdaptor, compiletime_point_tag>
	{
		// -------------------------------------------------------------------------------------- //

	public:

		// -------------------------------------------------------------------------------------- //

		typedef typename PointAdaptor::coordinate_type coordinate_type;

		typedef space::aabb<PointAdaptor> aabb_type;

		// -------------------------------------------------------------------------------------- //

		static void build(
			std::vector<PointAdaptor> &nodes,
			std::vector<PointAdaptor> &points,
			aabb_type &aabb,
			size_t node_index,
			size_t first_index,
			size_t last_index)
		{
			BOOST_STATIC_ASSERT(PointAdaptor::coordinate_count > 0);

			recursive_build<0>(nodes, points, aabb, node_index, first_index, last_index);
		}

		// -------------------------------------------------------------------------------------- //

	private:

		// -------------------------------------------------------------------------------------- //

		template <size_t A>
		static void recursive_build(
			std::vector<PointAdaptor> &nodes,
			std::vector<PointAdaptor> &points,
			aabb_type &aabb, size_t node_index,
			size_t first_index,
			size_t last_index)
		{
			SPACE_ASSERT( first_index < last_index , "current points range can't be empty, there must be a point which will be tree node" );
			SPACE_ASSERT( node_index < nodes.size() , "current node index can't be outside nodes recursive_container" );

			// find median point index for current 'zakres'
			size_t median_index = calculate_median_index(first_index, last_index);
			SPACE_ASSERT( median_index < points.size() , "point median index can't be outside points recursive_container" );

			// find n-th point
			std::nth_element(points.begin() + first_index, points.begin() + median_index, points.begin() + last_index, pt_comp_ct<A>());

			// create new node
			nodes[node_index] = *(points.begin() + median_index);

			static const size_t next_axis = (A + 1) % PointAdaptor::coordinate_count;

			// process left child node

			size_t left_node_index = 2 * node_index;
			// terminate - no more points
			if ( left_node_index >= nodes.size() )
				return;
			// create left node's aabb
			aabb_type left_aabb( aabb.template left_aabb<A>(nodes[node_index].template get<A>()) );
			// recursively call build for the left node
			recursive_build<next_axis>(nodes, points, left_aabb, left_node_index, first_index, median_index);

			// process right child node

			size_t right_node_index = left_node_index + 1;
			// terminate - no more points
			if ( right_node_index >= nodes.size() )
				return;
			// create right node's aabb
			aabb_type right_aabb( aabb.template right_aabb<A>(nodes[node_index].template get<A>()) );
			// recursively call build for the right node
			recursive_build<next_axis>(nodes, points, right_aabb, right_node_index, median_index + 1, last_index);
		}
	};

	// -------------------------------------------------------------------------------------- //

	class pt_comp
	{
	public:
		pt_comp(size_t axis)
			: m_axis(axis)
		{}

		template <typename Point>
		bool operator()(Point const& p1, Point const& p2) const
		{
			return p1.get(m_axis) < p2.get(m_axis);
		}
	private:
		size_t m_axis;
	};

	// -------------------------------------------------------------------------------------- //

	template <typename PointAdaptor>
	class kdtree_algorithms<PointAdaptor, runtime_point_tag>
	{
		// -------------------------------------------------------------------------------------- //

	public:

		// -------------------------------------------------------------------------------------- //

		typedef typename PointAdaptor::coordinate_type coordinate_type;
		typedef space::aabb<PointAdaptor> aabb_type;

		// -------------------------------------------------------------------------------------- //

		static void build(
			std::vector<PointAdaptor> &nodes,
			std::vector<PointAdaptor> &points,
			aabb_type &aabb,
			size_t node_index,
			size_t first_index,
			size_t last_index)
		{
			BOOST_STATIC_ASSERT(PointAdaptor::coordinate_count > 0);

			recursive_build(0, nodes, points, aabb, node_index, first_index, last_index);
		}

		// -------------------------------------------------------------------------------------- //

	private:

		// -------------------------------------------------------------------------------------- //

		static void recursive_build(
			size_t axis,
			std::vector<PointAdaptor> &nodes,
			std::vector<PointAdaptor> &points,
			aabb_type &aabb, size_t node_index,
			size_t first_index,
			size_t last_index)
		{
			SPACE_ASSERT( first_index < last_index , "current points range can't be empty, there must be a point which will be tree node" );
			SPACE_ASSERT( node_index < nodes.size() , "current node index can't be outside nodes recursive_container" );

			// find median point index for current 'zakres'
			size_t median_index = calculate_median_index(first_index, last_index);
			SPACE_ASSERT( median_index < points.size() , "point median index can't be outside points recursive_container" );

			// find n-th point
			std::nth_element(points.begin() + first_index, points.begin() + median_index, points.begin() + last_index, pt_comp(axis));

			// create new node
			nodes[node_index] = *(points.begin() + median_index);

			size_t next_axis = (axis + 1) % PointAdaptor::coordinate_count;

			// process left child node

			size_t left_node_index = 2 * node_index;
			// terminate - no more points
			if ( left_node_index >= nodes.size() )
				return;
			// create left node's aabb
			aabb_type left_aabb( aabb.left_aabb(axis, nodes[node_index].get(axis)) );
			// recursively call build for the left node
			recursive_build(next_axis, nodes, points, left_aabb, left_node_index, first_index, median_index);

			// process right child node

			size_t right_node_index = left_node_index + 1;
			// terminate - no more points
			if ( right_node_index >= nodes.size() )
				return;
			// create right node's aabb
			aabb_type right_aabb( aabb.right_aabb(axis, nodes[node_index].get(axis)) );
			// recursively call build for the right node
			recursive_build(next_axis, nodes, points, right_aabb, right_node_index, median_index + 1, last_index);
		}

		// -------------------------------------------------------------------------------------- //
	};

	// -------------------------------------------------------------------------------------- //

	template <typename PointAdaptor>
	class kdtree_algorithms<PointAdaptor, dynamic_point_tag>
	{
		// -------------------------------------------------------------------------------------- //

	public:

		// -------------------------------------------------------------------------------------- //

		typedef typename PointAdaptor::coordinate_type coordinate_type;
		typedef space::aabb<PointAdaptor> aabb_type;

		// -------------------------------------------------------------------------------------- //

		static void build(
			std::vector<PointAdaptor> &nodes,
			std::vector<PointAdaptor> &points,
			aabb_type &aabb,
			size_t node_index,
			size_t first_index,
			size_t last_index)
		{
			size_t coordinate_count = points[first_index].coordinate_count();
			SPACE_ASSERT(coordinate_count > 0, "points should have coordinates");
			recursive_build(0, nodes, points, aabb, node_index, first_index, last_index, coordinate_count);
		}

		// -------------------------------------------------------------------------------------- //

	private:

		// -------------------------------------------------------------------------------------- //

		static void recursive_build(
			size_t axis,
			std::vector<PointAdaptor> &nodes,
			std::vector<PointAdaptor> &points,
			aabb_type &aabb,
			size_t node_index,
			size_t first_index,
			size_t last_index,
			size_t coordinate_count)
		{
			SPACE_ASSERT( first_index < last_index , "current points range can't be empty, there must be a point which will be tree node" );
			SPACE_ASSERT( node_index < nodes.size() , "current node index can't be outside nodes recursive_container" );

			// find median point index for current 'zakres'
			size_t median_index = calculate_median_index(first_index, last_index);
			SPACE_ASSERT( median_index < points.size() , "point median index can't be outside points recursive_container" );

			// find n-th point
			std::nth_element(points.begin() + first_index, points.begin() + median_index, points.begin() + last_index, pt_comp(axis));

			// create new node
			nodes[node_index] = *(points.begin() + median_index);

			SPACE_ASSERT(points[first_index].coordinate_count() == coordinate_count, "wrong number of coordinates");

			size_t next_axis = (axis + 1) % coordinate_count;

			// process left child node

			size_t left_node_index = 2 * node_index;
			// terminate - no more points
			if ( left_node_index >= nodes.size() )
				return;
			// create left node's aabb
			aabb_type left_aabb( aabb.left_aabb(axis, nodes[node_index].get(axis)) );
			// recursively call build for the left node
			recursive_build(next_axis, nodes, points, left_aabb, left_node_index, first_index, median_index, coordinate_count);

			// process right child node

			size_t right_node_index = left_node_index + 1;
			// terminate - no more points
			if ( right_node_index >= nodes.size() )
				return;
			// create right node's aabb
			aabb_type right_aabb( aabb.right_aabb(axis, nodes[node_index].get(axis)) );
			// recursively call build for the right node
			recursive_build(next_axis, nodes, points, right_aabb, right_node_index, median_index + 1, last_index, coordinate_count);
		}

		// -------------------------------------------------------------------------------------- //
	};

	// -------------------------------------------------------------------------------------- //
}

// -------------------------------------------------------------------------------------- //

template <typename Point, typename PointAdaptor = typename point_traits<Point>::point_adaptor >
class kdtree
{
	// -------------------------------------------------------------------------------------- //

public:

	// -------------------------------------------------------------------------------------- //

	typedef Point point_type;
	typedef typename PointAdaptor::coordinate_type coordinate_type;

	typedef aabb<PointAdaptor> aabb_type;

	// -------------------------------------------------------------------------------------- //

	inline kdtree()
	{
	}

	// -------------------------------------------------------------------------------------- //

	template <typename Iter>
	inline kdtree(Iter first, Iter last)
	{
		build(first, last);
	}

	// -------------------------------------------------------------------------------------- //

	template <typename Iter>
	void build(Iter first, Iter last)
	{
		clear();

		if ( first == last )
			return;

		// calculate all object's aabb
		m_aabb = aabb_type(static_cast<PointAdaptor>(*first), static_cast<PointAdaptor>(*first));
		size_t points_number = 1;

		Iter iter = first;
		++iter;

		for( ; iter != last ; ++iter, ++points_number )
			m_aabb.expand(static_cast<PointAdaptor>(*iter));

		// resize nodes
		m_nodes.resize(points_number + 1);

		// create vector of points used in build process
		std::vector<PointAdaptor> points;
		points.reserve(points_number);
		std::copy(first, last, std::back_inserter(points));
		
		// create tree
		detail::kdtree_algorithms<PointAdaptor, typename PointAdaptor::point_category>::build(m_nodes, points, m_aabb, 1, 0, points_number);
	}

	// -------------------------------------------------------------------------------------- //

	inline void clear()
	{
		m_nodes.clear();
	}

	// -------------------------------------------------------------------------------------- //

	inline bool empty() const
	{
		return m_nodes.empty();
	}

	// -------------------------------------------------------------------------------------- //

	inline size_t size() const
	{
		return empty() ? 0 : m_nodes.size() - 1;
	}

	// -------------------------------------------------------------------------------------- //

private:

	// -------------------------------------------------------------------------------------- //

	std::vector<PointAdaptor> m_nodes;
	aabb_type m_aabb;

	// -------------------------------------------------------------------------------------- //
};

// -------------------------------------------------------------------------------------- //

}

// -------------------------------------------------------------------------------------- //

#endif // SPACE_SPAR_KDTREE_HPP
