// (C) Copyright Adam Wulkiewicz 2010

#ifndef SPACE_TYPES_HPP
#define SPACE_TYPES_HPP

// -------------------------------------------------------------------------------------- //

#include <boost/static_assert.hpp>

#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif
#include <limits>

#include "assert.hpp"

// -------------------------------------------------------------------------------------- //

namespace space
{

// -------------------------------------------------------------------------------------- //
// points adaptors
// -------------------------------------------------------------------------------------- //

struct compiletime_point_tag {};
struct runtime_point_tag {};
struct dynamic_point_tag {};

// -------------------------------------------------------------------------------------- //

template <typename Point>
struct compiletime_point : private Point
{
	typedef compiletime_point_tag point_category;
	typedef typename Point::coordinate_type coordinate_type;
	static const size_t coordinate_count = Point::coordinate_count;

	inline compiletime_point() {}
	inline compiletime_point(Point const& pt) : Point(pt) {}

	template <size_t K>	inline coordinate_type const & get() const { return Point::template get<K>(); }
	template <size_t K>	inline void set(coordinate_type const& v) { Point::template set<K>(v); }
};

// -------------------------------------------------------------------------------------- //

template <typename Point>
struct runtime_point : private Point
{
	typedef runtime_point_tag point_category;
	typedef typename Point::coordinate_type coordinate_type;
	static const size_t coordinate_count = Point::coordinate_count;

	inline runtime_point() {}
	inline runtime_point(Point const& pt) : Point(pt) {}

	inline coordinate_type const& get(size_t i) const { return Point::get(i); }
	inline void set(size_t i, coordinate_type const& v) { Point::set(i, v); }
};

// -------------------------------------------------------------------------------------- //

template <typename Point>
struct runtime_point_indexop : private Point
{
	typedef runtime_point_tag point_category;
	typedef typename Point::coordinate_type coordinate_type;
	static const size_t coordinate_count = Point::coordinate_count;

	inline runtime_point_indexop() {}
	inline runtime_point_indexop(Point const& pt) : Point(pt) {}

	inline coordinate_type const& get(size_t i) const { return Point::operator[](i); }
	inline void set(size_t i, coordinate_type const& v) { Point::operator[](i) = v; }
};

// -------------------------------------------------------------------------------------- //

template <typename Point>
struct runtime_point_array : private Point
{
	typedef runtime_point_tag point_category;
	typedef typename Point::value_type coordinate_type;
	static const size_t coordinate_count = Point::static_size;
	
	inline runtime_point_array() {}
	inline runtime_point_array(Point const& pt) : Point(pt) {}

	inline coordinate_type const& get(size_t i) const { return Point::operator[](i); }
	inline void set(size_t i, coordinate_type const& v) { Point::operator[](i) = v; }
};

// -------------------------------------------------------------------------------------- //

template <typename Point>
struct dynamic_point : private Point
{
	typedef dynamic_point_tag point_category;
	typedef typename Point::coordinate_type coordinate_type;
	
	inline dynamic_point() {}
	inline dynamic_point(Point const& pt) : Point(pt) {}

	inline size_t coordinate_count() const { return Point::coordinate_count(); }

	inline coordinate_type const& get(size_t i) const { return Point::get(i); }
	inline void set(size_t i, coordinate_type const& v) { Point::set(i, v); }
};

// -------------------------------------------------------------------------------------- //

template <typename Point>
struct dynamic_point_randcont : private Point
{
	typedef dynamic_point_tag point_category;
	typedef typename Point::value_type coordinate_type;
	
	inline dynamic_point_randcont() {}
	inline dynamic_point_randcont(Point const& pt) : Point(pt) {}

	inline size_t coordinate_count() const { return Point::size(); }

	inline coordinate_type const& get(size_t i) const { return Point::operator[](i); }
	inline void set(size_t i, coordinate_type const& v) { Point::operator[](i) = v; }
};

// -------------------------------------------------------------------------------------- //
// point traits
// -------------------------------------------------------------------------------------- //

template <typename Point>
struct point_traits
{
	typedef runtime_point_indexop<Point> point_adaptor;
};

// -------------------------------------------------------------------------------------- //
// conversion adaptors
// -------------------------------------------------------------------------------------- //

template <typename Point>
struct compiletime_from_runtime_point : private Point
{
	typedef compiletime_point_tag point_category;
	typedef typename Point::coordinate_type coordinate_type;
	static const size_t coordinate_count = Point::coordinate_count;

	inline compiletime_from_runtime_point() {}
	inline compiletime_from_runtime_point(Point const& pt) : Point(pt) {}

	template <size_t K>	inline coordinate_type const & get() const { return Point::get(K); }
	template <size_t K>	inline void set(coordinate_type const& v) { Point::set(K, v); }
};

// -------------------------------------------------------------------------------------- //

namespace detail
{
	// -------------------------------------------------------------------------------------- //

	template <size_t A>
	struct runtime_from_compiletime_point_algorithms
	{
		template <typename Point>
		inline static typename Point::coordinate_type const& get(Point const& pt, size_t i)
		{
			if ( i == A )
				return pt.template get<A>();
			else
				return runtime_from_compiletime_point_algorithms<A - 1>::get(pt, i);
		}

		// -------------------------------------------------------------------------------------- //

		template <typename Point>
		inline static void set(Point & pt, size_t i, typename Point::coordinate_type const& v)
		{
			if ( i == A )
			{
				pt.template set<A>(v);
				return;
			}
			else
				runtime_from_compiletime_point_algorithms<A - 1>::set(pt, i, v);
		}
	};

	// -------------------------------------------------------------------------------------- //

	template <>
	struct runtime_from_compiletime_point_algorithms<0>
	{
		template <typename Point>
		inline static typename Point::coordinate_type const& get(Point const& pt, size_t i)
		{
			SPACE_ASSERT(i == 0, "i should be equal to 0");
			return pt.template get<0>();
		}

		// -------------------------------------------------------------------------------------- //

		template <typename Point>
		inline static void set(Point & pt, size_t i, typename Point::coordinate_type const& v)
		{
			SPACE_ASSERT(i == 0, "i should be equal to 0");
			pt.template set<0>(v);
		}
	};

	// -------------------------------------------------------------------------------------- //
}

// -------------------------------------------------------------------------------------- //

template <typename Point>
struct runtime_from_compiletime_point : private Point
{
	typedef runtime_point_tag point_category;
	typedef typename Point::coordinate_type coordinate_type;
	static const size_t coordinate_count = Point::coordinate_count;

	inline runtime_from_compiletime_point() {}
	inline runtime_from_compiletime_point(Point const& pt) : Point(pt) {}

	inline coordinate_type const& get(size_t i) const { return detail::runtime_from_compiletime_point_algorithms<coordinate_count - 1>::get(static_cast<Point const&>(*this), i); }
	inline void set(size_t i, coordinate_type const& v) { return detail::runtime_from_compiletime_point_algorithms<coordinate_count - 1>::set(static_cast<Point &>(*this), i, v); }
};

// -------------------------------------------------------------------------------------- //
// aabb
// -------------------------------------------------------------------------------------- //

namespace detail
{
	// -------------------------------------------------------------------------------------- //

	template <typename Point>
	class aabb_base
	{
		// -------------------------------------------------------------------------------------- //

	public:

		// -------------------------------------------------------------------------------------- //

		inline aabb_base()
		{}
		inline aabb_base(Point const& min_corner, Point const& max_corner)
			: m_min_corner(min_corner), m_max_corner(max_corner)
		{}

		inline Point const& min_corner() const { return m_min_corner; }
		inline Point const& max_corner() const { return m_max_corner; }
		inline Point& min_corner() { return m_min_corner; }
		inline Point& max_corner() { return m_max_corner; }

		// -------------------------------------------------------------------------------------- //

	private:

		// -------------------------------------------------------------------------------------- //

		Point m_min_corner;
		Point m_max_corner;

		// -------------------------------------------------------------------------------------- //
	};

	// -------------------------------------------------------------------------------------- //

	template <typename PointAdapter, typename PointCategory>
	struct aabb
	{};

	// -------------------------------------------------------------------------------------- //

	template <size_t A>
	struct aabb_algorithms_ct
	{
		// -------------------------------------------------------------------------------------- //

		template <typename PointCt>
		inline static bool good(PointCt const& min_corner, PointCt const& max_corner)
		{
			return min_corner.template get<A>() <= max_corner.template get<A>()
				&& aabb_algorithms_ct<A - 1>::good(min_corner, max_corner);
		}

		// -------------------------------------------------------------------------------------- //

		template <typename PointCt>
		inline static void prepare_to_expand(PointCt const& min_corner, PointCt const& max_corner)
		{
			min_corner.template set<A>(std::numeric_limits<typename PointCt::coordinate_type>::max());
			max_corner.template set<A>(-std::numeric_limits<typename PointCt::coordinate_type>::max());
			aabb_algorithms_ct<A - 1>::prepare_to_expand(min_corner, max_corner);
		}

		// -------------------------------------------------------------------------------------- //

		template <typename PointCt>
		inline static void expand(PointCt & min_corner, PointCt & max_corner, PointCt const& min_corner2, PointCt const& max_corner2)
		{
			SPACE_ASSERT(min_corner2.template get<A>() <= max_corner2.template get<A>(), "min must be <= max");

			if ( min_corner2.template get<A>() < min_corner.template get<A>() )
				min_corner.template set<A>(min_corner2.template get<A>());
			if ( max_corner2.template get<A>() > max_corner.template get<A>() )
				max_corner.template set<A>(max_corner2.template get<A>());

			aabb_algorithms_ct<A - 1>::expand(min_corner, max_corner, min_corner2, max_corner2);
		}

		// -------------------------------------------------------------------------------------- //
	};

	// -------------------------------------------------------------------------------------- //

	template <>
	struct aabb_algorithms_ct<0>
	{
		// -------------------------------------------------------------------------------------- //

		template <typename PointCt>
		inline static bool good(PointCt const& min_corner, PointCt const& max_corner)
		{
			return min_corner.template get<0>() <= max_corner.template get<0>();
		}

		// -------------------------------------------------------------------------------------- //

		template <typename PointCt>
		inline static void prepare_to_expand(PointCt const& min_corner, PointCt const& max_corner)
		{
			min_corner.template set<0>(std::numeric_limits<typename PointCt::coordinate_type>::max());
			max_corner.template set<0>(-std::numeric_limits<typename PointCt::coordinate_type>::max());
		}

		// -------------------------------------------------------------------------------------- //

		template <typename PointCt>
		inline static void expand(PointCt & min_corner, PointCt & max_corner, PointCt const& min_corner2, PointCt const& max_corner2)
		{
			SPACE_ASSERT(min_corner2.template get<0>() <= max_corner2.template get<0>(), "min must be <= max");

			if ( min_corner2.template get<0>() < min_corner.template get<0>() )
				min_corner.template set<0>(min_corner2.template get<0>());
			if ( max_corner2.template get<0>() > max_corner.template get<0>() )
				max_corner.template set<0>(max_corner2.template get<0>());
		}

		// -------------------------------------------------------------------------------------- //
	};

	// -------------------------------------------------------------------------------------- //

	template <typename PointAdapter>
	struct aabb<PointAdapter, compiletime_point_tag> : public aabb_base<PointAdapter>
	{
		// -------------------------------------------------------------------------------------- //

		typedef typename PointAdapter::coordinate_type coordinate_type;

		BOOST_STATIC_ASSERT(PointAdapter::coordinate_count > 0);

		// -------------------------------------------------------------------------------------- //

		inline aabb()
			: aabb_base<PointAdapter>()
		{}
		inline aabb(PointAdapter const& min_corner, PointAdapter const& max_corner)
			: aabb_base<PointAdapter>(min_corner, max_corner)
		{}

		using aabb_base<PointAdapter>::min_corner;
		using aabb_base<PointAdapter>::max_corner;

		// -------------------------------------------------------------------------------------- //

		inline bool good()
		{
			return aabb_algorithms_ct<PointAdapter::coordinate_count - 1>::good(min_corner(), max_corner());
		}

		// -------------------------------------------------------------------------------------- //

		inline void prepare_to_expand()
		{
			aabb_algorithms_ct<PointAdapter::coordinate_count - 1>::prepare_to_expand(min_corner(), max_corner());
		}

		// -------------------------------------------------------------------------------------- //

		inline void expand(aabb const& box)
		{
			aabb_algorithms_ct<PointAdapter::coordinate_count - 1>::expand(min_corner(), max_corner(), box.min_corner(), box.max_corner());
		}

		// -------------------------------------------------------------------------------------- //

		inline void expand(PointAdapter const& pt)
		{
			aabb_algorithms_ct<PointAdapter::coordinate_count - 1>::expand(min_corner(), max_corner(), pt, pt);
		}

		// -------------------------------------------------------------------------------------- //

		template <size_t A>
		inline aabb left_aabb(coordinate_type const& plane_position) const
		{
			BOOST_STATIC_ASSERT(A < PointAdapter::coordinate_count);

			SPACE_ASSERT(min_corner().template get<A>() <= max_corner().template get<A>(), "min must be <= max");
			SPACE_ASSERT(min_corner().template get<A>() <= plane_position && plane_position <= max_corner().template get<A>(), "wrong plane's position");

			PointAdapter new_max_corner = max_corner();
			new_max_corner.template set<A>(plane_position);

			return aabb(min_corner(), new_max_corner);
		}

		// -------------------------------------------------------------------------------------- //

		template <size_t A>
		inline aabb right_aabb(coordinate_type const& plane_position) const
		{
			BOOST_STATIC_ASSERT(A < PointAdapter::coordinate_count);

			SPACE_ASSERT(min_corner().template get<A>() <= max_corner().template get<A>(), "min must be <= max");
			SPACE_ASSERT(min_corner().template get<A>() <= plane_position && plane_position <= max_corner().template get<A>(), "wrong plane's position");

			PointAdapter new_min_corner = min_corner();
			new_min_corner.template set<A>(plane_position);

			return aabb(new_min_corner, max_corner());
		}

		// -------------------------------------------------------------------------------------- //
	};

	// -------------------------------------------------------------------------------------- //

	struct aabb_algorithms_rt_d
	{
		// -------------------------------------------------------------------------------------- //

		template <typename PointAdapter>
		inline static bool good(PointAdapter const& min_corner, PointAdapter const& max_corner, size_t coordinate_count)
		{
			for ( size_t a = 0 ; a < coordinate_count ; ++a )
			{
				if ( min_corner.get(a) > max_corner.get(a) )
					return false;
			}
			return true;
		}

		// -------------------------------------------------------------------------------------- //

		template <typename PointAdapter>
		inline void prepare_to_expand(PointAdapter const& min_corner, PointAdapter const& max_corner, size_t coordinate_count)
		{
			for ( size_t a = 0 ; a < coordinate_count ; ++a )
			{
				min_corner.get(a) = std::numeric_limits<typename PointAdapter::coordinate_type>::max();
				max_corner.get(a) = -std::numeric_limits<typename PointAdapter::coordinate_type>::max();
			}
		}

		// -------------------------------------------------------------------------------------- //

		template <typename PointAdapter>
		inline static void expand(PointAdapter & min_corner, PointAdapter & max_corner, PointAdapter const& min_corner2, PointAdapter const& max_corner2, size_t coordinate_count)
		{
			for ( size_t a = 0 ; a < coordinate_count ; ++a )
			{
				SPACE_ASSERT(min_corner2.get(a) <= max_corner2.get(a), "min must be <= max");

				if ( min_corner2.get(a) < min_corner.get(a) )
					min_corner.set(a, min_corner2.get(a));
				if ( max_corner2.get(a) > max_corner.get(a) )
					max_corner.set(a, max_corner2.get(a));
			}
		}

		// -------------------------------------------------------------------------------------- //

		template <typename PointAdapter>
		inline static aabb<PointAdapter, typename PointAdapter::point_category> left_aabb(
			PointAdapter const& min_corner,
			PointAdapter const& max_corner,
			size_t coordinate_count,
			size_t axis,
			typename PointAdapter::coordinate_type const& plane_position)
		{
			SPACE_ASSERT(axis < coordinate_count, "axis too big");
			SPACE_ASSERT(min_corner.get(axis) <= max_corner.get(axis), "min must be <= max");
			SPACE_ASSERT(min_corner.get(axis) <= plane_position && plane_position <= max_corner.get(axis), "wrong plane's position");

			PointAdapter new_max_corner = max_corner;
			new_max_corner.set(axis, plane_position);

			return aabb<PointAdapter, typename PointAdapter::point_category>(min_corner, new_max_corner);
		}

		// -------------------------------------------------------------------------------------- //

		template <typename PointAdapter>
		inline static aabb<PointAdapter, typename PointAdapter::point_category> right_aabb(
			PointAdapter const& min_corner,
			PointAdapter const& max_corner,
			size_t coordinate_count,
			size_t axis,
			typename PointAdapter::coordinate_type const& plane_position)
		{
			SPACE_ASSERT(axis < coordinate_count, "axis too big");
			SPACE_ASSERT(min_corner.get(axis) <= max_corner.get(axis), "min must be <= max");
			SPACE_ASSERT(min_corner.get(axis) <= plane_position && plane_position <= max_corner.get(axis), "wrong plane's position");

			PointAdapter new_min_corner = min_corner;
			new_min_corner.set(axis, plane_position);

			return aabb<PointAdapter, typename PointAdapter::point_category>(new_min_corner, max_corner);
		}

		// -------------------------------------------------------------------------------------- //
	};

	// -------------------------------------------------------------------------------------- //

	template <typename PointAdapter>
	struct aabb<PointAdapter, runtime_point_tag> : public aabb_base<PointAdapter>
	{
		// -------------------------------------------------------------------------------------- //

		typedef typename PointAdapter::coordinate_type coordinate_type;

		// -------------------------------------------------------------------------------------- //

		inline aabb()
			: aabb_base<PointAdapter>()
		{}
		inline aabb(PointAdapter const& min_corner, PointAdapter const& max_corner)
			: aabb_base<PointAdapter>(min_corner, max_corner)
		{}

		using aabb_base<PointAdapter>::min_corner;
		using aabb_base<PointAdapter>::max_corner;

		// -------------------------------------------------------------------------------------- //

		inline bool good() const
		{
			return aabb_algorithms_rt_d::good(min_corner(), max_corner(), PointAdapter::coordinate_count);
		}

		// -------------------------------------------------------------------------------------- //

		inline void prepare_to_expand()
		{
			aabb_algorithms_rt_d::prepare_to_expand(min_corner(), max_corner(), PointAdapter::coordinate_count);
		}

		// -------------------------------------------------------------------------------------- //

		inline void expand(aabb const& box)
		{
			aabb_algorithms_rt_d::expand(min_corner(), max_corner(), box.min_corner(), box.max_corner(), PointAdapter::coordinate_count);
		}

		// -------------------------------------------------------------------------------------- //

		inline void expand(PointAdapter const& point)
		{
			aabb_algorithms_rt_d::expand(min_corner(), max_corner(), point, point, PointAdapter::coordinate_count);
		}

		// -------------------------------------------------------------------------------------- //

		inline aabb left_aabb(size_t axis, coordinate_type const& plane_position) const
		{
			return aabb_algorithms_rt_d::left_aabb(min_corner(), max_corner(), PointAdapter::coordinate_count, axis, plane_position);
		}

		// -------------------------------------------------------------------------------------- //

		inline aabb right_aabb(size_t axis, coordinate_type const& plane_position) const
		{
			return aabb_algorithms_rt_d::right_aabb(min_corner(), max_corner(), PointAdapter::coordinate_count, axis, plane_position);
		}

		// -------------------------------------------------------------------------------------- //
	};

	// -------------------------------------------------------------------------------------- //

	template <typename PointAdapter>
	struct aabb<PointAdapter, dynamic_point_tag> : public aabb_base<PointAdapter>
	{
		// -------------------------------------------------------------------------------------- //

		typedef typename PointAdapter::coordinate_type coordinate_type;

		// -------------------------------------------------------------------------------------- //

		inline aabb()
			: aabb_base<PointAdapter>()
		{}
		inline aabb(PointAdapter const& min_corner, PointAdapter const& max_corner)
			: aabb_base<PointAdapter>(min_corner, max_corner)
		{
			SPACE_ASSERT(min_corner.coordinate_count() == max_corner.coordinate_count(), "coordinates number should be equal");
		}

		using aabb_base<PointAdapter>::min_corner;
		using aabb_base<PointAdapter>::max_corner;

		// -------------------------------------------------------------------------------------- //

		inline bool good() const
		{
			SPACE_ASSERT(min_corner().coordinate_count() == max_corner().coordinate_count(), "coordinates number should be equal");
			return aabb_algorithms_rt_d::good(min_corner(), max_corner(), min_corner().coordinate_count());
		}

		// -------------------------------------------------------------------------------------- //

		inline void prepare_to_expand()
		{
			SPACE_ASSERT(min_corner().coordinate_count() == max_corner().coordinate_count(), "coordinates number should be equal");
			aabb_algorithms_rt_d::prepare_to_expand(min_corner(), max_corner(), min_corner().coordinate_count());
		}

		// -------------------------------------------------------------------------------------- //

		inline void expand(aabb const& box)
		{
			SPACE_ASSERT(min_corner().coordinate_count() == max_corner().coordinate_count(), "coordinates number should be equal");
			aabb_algorithms_rt_d::expand(min_corner(), max_corner(), box.min_corner(), box.max_corner(), min_corner().coordinate_count());
		}

		// -------------------------------------------------------------------------------------- //

		inline void expand(PointAdapter const& point)
		{
			SPACE_ASSERT(min_corner().coordinate_count() == max_corner().coordinate_count(), "coordinates number should be equal");
			aabb_algorithms_rt_d::expand(min_corner(), max_corner(), point, point, min_corner().coordinate_count());
		}

		// -------------------------------------------------------------------------------------- //

		inline aabb left_aabb(size_t axis, coordinate_type const& plane_position) const
		{
			SPACE_ASSERT(min_corner().coordinate_count() == max_corner().coordinate_count(), "coordinates number should be equal");
			return aabb_algorithms_rt_d::left_aabb(min_corner(), max_corner(), min_corner().coordinate_count(), axis, plane_position);
		}

		// -------------------------------------------------------------------------------------- //

		inline aabb right_aabb(size_t axis, coordinate_type const& plane_position) const
		{
			SPACE_ASSERT(min_corner().coordinate_count() == max_corner().coordinate_count(), "coordinates number should be equal");
			return aabb_algorithms_rt_d::right_aabb(min_corner(), max_corner(), min_corner().coordinate_count(), axis, plane_position);
		}

		// -------------------------------------------------------------------------------------- //
	};

	// -------------------------------------------------------------------------------------- //
}

// -------------------------------------------------------------------------------------- //

template <typename PointAdaptor>
struct aabb : public detail::aabb<PointAdaptor, typename PointAdaptor::point_category>
{
	// -------------------------------------------------------------------------------------- //

	inline aabb()
		: detail::aabb<PointAdaptor, typename PointAdaptor::point_category>()
	{}
	inline aabb(PointAdaptor const& min_corner, PointAdaptor const& max_corner)
		: detail::aabb<PointAdaptor, typename PointAdaptor::point_category>(min_corner, max_corner)
	{}
	inline aabb(detail::aabb<PointAdaptor, typename PointAdaptor::point_category> const& base)
		: detail::aabb<PointAdaptor, typename PointAdaptor::point_category>(base)
	{}

	// -------------------------------------------------------------------------------------- //
};

// -------------------------------------------------------------------------------------- //

}

// -------------------------------------------------------------------------------------- //

#endif // SPACE_TYPES_HPP
